/*--------------------------------------------------------------------------*
 * readarkeia.c  ---  V5  ---  For ARKEIA  ----
 * Copyright (c) 1999-2002 ARKEIA
 * All rights reserved
 * Your licence agreement specifies the terms and conditions
 * for the use of this software
 *--------------------------------------------------------------------------*/

/* NL_NOUPDATE */

const char readarkeia_version[] = "5.0.14";

/*--------------------------------------------------------------------------*
 * IMPORTANT NOTICE: The following program is provided by ARKEIA for emergency
 * case only. ARKEIA denies all warranties about the safety of this program for
 * data existing on the machine. No implicit or explicit promise is made by
 * ARKEIA about the quality, the result or conformity with user needs.
 * The use of this source program is made under the user supervision and under
 * his sole responsibility.
 *
 * ARKEIA DENIES ALL RESPONSIBILITY FOR USE OF THIS PROGRAM AND/OR ITS RESULTS.
 *--------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*
 * READARKEIA: This program allows the restoration of backup tapes made with
 *          ARKEIA.
 *--------------------------------------------------------------------------*
 *
 * You can use readarkeia like it was a tar command. It's the same syntax.
 * Not all options are implemented.
 *
 * SYNOPSIS
 *   readarkeia [-l<loglevel>] x|t|i [mfuv] [device|-] [filename1 filename2 ...]
 *
 * DESCRIPTION
 *     -l   Change the verbose level. You can use it to obtain detailled
 *          informations on what is really on the tape. This value can
 *          be from 10 to 80.
 *
 *     t    List the table of contents of the tape.
 *
 *     x    Extract the named files from the tape. If no files are specified
 *          then all the tape is restored.
 *
 *     i    Get informations about the tape.
 *
 *     m    Ask to readarkeia the we are in the middle of the tape.
 *          (There is no label to read).
 *
 *     f    Use the next argument as the name of the device. If device name
 *          is given as `-', readarkeia reads from the  standard  input
 *          the tape datas.
 *
 *     u    Update precedent versions of each files. Normally, when readarkeia
 *          find different versions of the same file (named 'myfile' for
 *          example) on tapes, it renames them 'myfile.1', 'myfile.2', etc.
 *          With this option, only the more recent version of the file is
 *          restored, simply named 'myfile'.
 *
 *     v    Verbose.  Normally readarkeia  does  its  work  silently.
 *
 *     The filenames are optionnal. Has sayd before, If no files are specified
 *     then all the tape is restored. The filenames have the following syntax:
 *                <HOST>/<FILE_PATH or PATH_PART>
 *
 *     If you want to restore a complete path ore many path you can specify
 *     the common beginning. Example:
 *                The filename 'alpha/u' will math for path like
 *                    'alpha/usr', 'alpha/u3', 'alpha/users', ...
 *
 * EXAMPLES
 *     To get a table of contents of a tape put in the device '/dev/rmt0'
 *     on an AIX system. The command is:
 *              readarkeia tvf /dev/rmt0
 *
 *     Then if you want to extract file like '/usr/lib/libc.a' you will
 *     indicate at the beginning of the filename the hostname. So the filename
 *     will be: '<HOST>/usr/lib/libc.a'
 *     The command id:
 *              readarkeia xvf /dev/rmt0 alpha/usr/lib/libc.a
 *
 * WARNING
 *     readarkeia program will restore the last version of a file, renaming
 *     silently any existing version.
 *--------------------------------------------------------------------------*/

#include <stdio.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <malloc.h>
#include <stdlib.h>
#include <string.h>
#include <utime.h>

/*--------------------------------------------------------------------------*/

#define INT4	int
#define UBYTE	unsigned char	/* Unsigned byte */
#define UCHAR	unsigned char
#define UINT4	unsigned int
#define UINT	unsigned int
#define ULONG	unsigned long	/* Unsigned word (4 bytes) */
#define UWORD	unsigned short	/* Unsigned word (2 bytes) */
#define UCARD	unsigned	/* Which ever is fastest, uns. or int  */

/*--------------------------------------------------------------------------*/

#define STR_SIZE		1024
#define OUI			(-1)
#define NON			0

#ifndef MAX
# define MAX(a,b) ((a > b) ? a : b)
#endif

static int nl_loglevel = 0;	/* Default */
static int nl_line;

#define GOTO_ERROR		{nl_line=__LINE__; goto error;}
#define NL_LOG(level)		if(nl_loglevel>=(level)) printf
#define INFO(dummylevel)	if(global_operation==ADMIN_INFO) printf
#define NL_ERR(dummylevel)	\
	{ \
		 int local_errno; \
		 local_errno = errno; \
		 printf("ERROR: "); \
		 if (local_errno) { \
			 if ((local_errno > 0) && (local_errno < sys_nerr)) { \
				 printf("(%d: %s) ", local_errno, \
						 strerror(local_errno)); \
			 } else printf("(%d) ", local_errno); \
		 } \
	 } \
	 printf

/*--------------------------------------------------------------------------*/

extern int errno;
extern int sys_nerr;

char *compress_type();
int hextoi();
void end_of_file();
void myexit();
void uncompress();
void usage();

/*--------------------------------------------------------------------------*/

#define F_ST			252		/* strategy process */
#define HEADER_SIZE		12
#define MAXFLOWS		200		/* max number of flows */
#define OPB_SERV_VERSION	"@(#)OPB-SVERSION"
#define OPB_VERSION		"@(#)OPB-VERSION!!"
#define TAPEPAGESIZE		(8*1024)

#define INVALID_FD	(-1)
#define ARK_HANDLE	int
#define CLOSE(_M)	close(_M);

/*--------------------------------------------------------------------------*
 * Structure of a tape block
 * a seq number is 4 bits long, a number of 8k pages is 3 bits
 *--------------------------------------------------------------------------*/

#define BLKSZMASK		7	/* 00000111 */
#define DB_BLOCK		8	/* 00001000 */
#define O3MAGICTAPE	0xec557948

struct stapblk {
	UINT4 blkseqnum;	/* flow[8] + seqnum */
	UINT4 blkbkid;		/* backup id */
	UCHAR blkdummy;		/* 8 spare bits */
	UCHAR blksz;	/* (spare bits <<4) | DB_BLOCK | Number of 8k pages */
	UCHAR blkproffh; /* M_BKDAT: high byte offset of first prop on block */
			/* M_DBDAT: 0xff if 1st block of db for slot, else 0 */
	UCHAR blkproffl; /* M_BKDAT: low byte offset of first prop on block */
			/* M_DBDAT: 0xff if last block of db for slot, else 0*/
	UCHAR blkdata[1];	/* The block data */
};

/* Macro to extract flow number from a composite number
   ----------------------------------------------------*/
#define COMP2FLOW(num) ((num>>24)&255)

struct stapblk *in_blk = 0;

/*--------------------------------------------------------------------------*/

struct st_flow {
	int	 fnum;
	UINT4	 bkid;
	UINT4	 blknum;
	char	*prop;
	char	*value;
	char	*realsize;
	int	 state;	/* For prop reading */
	int	 operation;
	int	 bin_rest;

	/* Prop storage */
	char	*fullname;
	char	*host;
	char	*volume;
	char	*plugin;
	char	*location;
	char	*component;
	char	*link_path;
	char	*obj_type;
	char	*slave_version;
	mode_t	 type;
	dev_t	 ark_dev;	/* device file resides on */
	ino_t	 ark_ino;	/* the file serial number */
	mode_t	 ark_mode;	/* file mode */
	nlink_t	 ark_nlink;	/* number of hard links to the file */
	uid_t	 ark_uid;	/* user ID of owner */
	gid_t	 ark_gid;	/* group ID of owner */
	dev_t	 ark_rdev;	/* the device identifier (special files only)*/
	off_t	 ark_size;	/* total size of file, in bytes */
	time_t	 ark_atime;	/* file last access time */
	time_t	 ark_mtime;	/* file last modify time */
	time_t	 ark_ctime;	/* file last status change time */

	/* File section */
	char	*buff;
	int	 buff_size;
	int	 flags;			/* Bit mask */
	char	*cur_pc;
	ARK_HANDLE	  fd;
};

/* Defines for <state>
   -------------------- */
#define NORMAL_STATE		1
#define INCOMPLETE_PROP		2
#define INCOMPLETE_VAL		3
#define INCOMPLETE_REALSZ	4
#define INCOMPLETE_BIN		5

/* Defines for <operation>
   ----------------------- */
#define NO_OPERATION		1
#define EXTRACT_FILE		2
#define DONT_EXTRACT		3
#define CONSULT_LIST		4
#define ADMIN_INFO		5

/* Defines for <type>
   ------------------ */
#define OPB_BLKDEV		1
#define OPB_CHRDEV		2
#define OPB_DIR			3
#define OPB_FIFO		4
#define OPB_NAMEDFILE		5
#define OPB_OBJECT		6
#define OPB_RAWDEV		7
#define OPB_REGFILE		8
#define OPB_SYMLINK		9
#define OPB_UNKNOWN		10

/* Compression values */
#define COMPRESS_MASK		3
#define NO_COMPRESS		0
#define LZ1_COMPRESS		1
#define LZ3_3_COMPRESS		2

/* Defines for <flags>
   ------------------- */
#define COMPRESS_BUFF	(1<<1)
#define DONT_FREE_BUFF	(1<<2)
#define HARD_LINK	(1<<3)

/* Defines for <fd>
   ---------------- */
#define INVALID_FD	(-1)
static int fd_in = INVALID_FD;

struct st_flow flow_tab[MAXFLOWS];

static int global_operation = NO_OPERATION;
static char *extract_path[1024];
static int nb_path = 0;

static short gl_update_mode = NON;  /* set to OUI if the 'u' option is used */

/*--------------------------------------------------------------------------*/

#define GETWORD(spc, val, mpc) \
	pc = getword(spc, val, mpc); \
	if (!pc) goto unexp_eob; \
	pc++

#define GETNCHARS(spc, val, mpc, n) \
	pc = getnchars(spc, val, mpc, n); \
	if (!pc) goto unexp_eob;

#define ALLOC_IF_NOT(ptr) \
	if (!(ptr)) { \
		if (!((ptr) = malloc(STR_SIZE+1))) { \
			NL_ERR(0)("Can't alloc memory for prop storage\n"); \
			GOTO_ERROR; \
		} \
		*(ptr) = 0; \
	 }

/*--------------------------------------------------------------------------*/

static int need_uint4_chsex = NON;

UINT4 ch_uint4_sex(in)
UINT4 in;
{
	UINT4 out;
	char *p1, *p2;

	need_uint4_chsex = OUI;
	p1 = (char *)&in;
	p2 = (char *)&out;
	p2[0] = p1[3];
	p2[1] = p1[2];
	p2[2] = p1[1];
	p2[3] = p1[0];
	return (out);
}

/*--------------------------------------------------------------------------*/

int match(filename)
char *filename;
{
	int i;
	char *pc1, *pc2;

	for (i = 0 ; i < nb_path ; i++) {
		for (pc1 = filename, pc2 = extract_path[i] ;
		     *pc1 && *pc2 && (*pc1 == *pc2); pc1++, pc2++) ;
		if ((!*pc1 && !*pc2) || (*pc1 && !*pc2)) return (OUI);
	}
	return (NON);
}

/*--------------------------------------------------------------------------*/

void check_operation(pflow)
struct st_flow *pflow;
{
	char *pc1, *pc2;
	struct stat buf;
	int i;
	char str[1024];

	/* Make the full file name */
	pc2 = pflow->fullname;
	for (pc1 = pflow->host ; *pc1 ; *pc2 = *pc1, pc1++, pc2++) ;
	if (*pflow->plugin) {
		if (*(pc2 - 1) != '/') *pc2++ = '/';
		for (pc1 = pflow->plugin; *pc1; *pc2 = *pc1, pc1++, pc2++) ;
	}
	if (*pflow->volume) {
		if (*(pc2 - 1) != '/') *pc2++ = '/';
		for (pc1 = pflow->volume ; *pc1 ; *pc2 = *pc1, pc1++, pc2++) ;
	}
	if (*pflow->location != '/') *pc2++ = '/';
	if ((strlen(pflow->location) == 1) &&
	   (pflow->location[0] == '/')) goto skip_location;
	for (pc1 = pflow->location ; *pc1 ; *pc2 = *pc1, pc1++, pc2++) ;
skip_location:
	if (*(pc2 - 1) != '/') *pc2++ = '/';
	for (pc1 = pflow->component ; *pc1 ; *pc2 = *pc1, pc1++, pc2++) ;
	if (*(pc2 - 1) == '/') pc2--;
	*pc2 = '\0';

	if (!*pflow->location || !*pflow->host || !*pflow->fullname) {
		pflow->operation = DONT_EXTRACT;
		goto its_ok;
	}

	pflow->operation = global_operation;
	if ((global_operation != EXTRACT_FILE) || !nb_path) goto its_ok;

	/* EXTRACT_FILE */
	if (!match(pflow->fullname)) pflow->operation = DONT_EXTRACT;

its_ok:
	if (pflow->operation == EXTRACT_FILE) {
		/* Check if this file already exist */
		if (stat(pflow->fullname, &buf) == 0 && !S_ISDIR(buf.st_mode)) {
			if (gl_update_mode == NON) {
				i = 0;
				while (++i) {
					sprintf(str, "%s.%d", pflow->fullname, i);
					if (stat(str, &buf)) {
						rename(pflow->fullname, str);
						break ;
					}
				}
			} else if (buf.st_mtime > pflow->ark_mtime) {
				pflow->operation = DONT_EXTRACT;
			}
		}
	}
	return ;

}

/*--------------------------------------------------------------------------*
 * set access and modification times
 *--------------------------------------------------------------------------*/
static void set_time(namep, atime, mtime)
register char *namep;
time_t atime, mtime;
{
	struct utimbuf times;

	times.actime = atime;
	times.modtime = mtime;
	utime(namep, &times);
}

/*--------------------------------------------------------------------------*
 * create recursively all dirs up to and including name
 *--------------------------------------------------------------------------*/

int /*NL_DCL*/gendirs(name,including)
char *name;
int including; /* 0: do not create name, 1: create dir name */
{
	char *pc,*pc1;
	char str[1024];
	struct stat sta;
	static int recurs = -1;
	int domk;

	if (!*name) GOTO_ERROR;

	recurs++;
	if (!including) {
		strcpy(str, name);
		for (pc = pc1 = str; *pc; pc++) if (*pc == '/') pc1 = pc;
		*pc1 = 0;
		if (!gendirs(str,1)) GOTO_ERROR;
		goto retoui;
	}
	domk = 0;
	if (stat(name,&sta) < 0) domk = 1;
	else if (S_ISREG(sta.st_mode)) {
		if (unlink(name) < 0) {
			NL_ERR(0)("Can't remove '%s'\n",name);
			GOTO_ERROR;
		}
		domk = 1;
	}
	if (domk) {
		if (mkdir(name, 0755) < 0) {
			if (!gendirs(name, 0)) GOTO_ERROR;
			if (mkdir(name, 0755) < 0) {
				NL_ERR(0)("Can't create directory '%s'\n", name);
				GOTO_ERROR;
			}
		}
	}
retoui:
	recurs--;
	return (OUI);
error:
	NL_ERR(0)("Error for directory '%s'\n", name);
	recurs--;
	return (NON);
}

/*--------------------------------------------------------------------------*/

int create_file(pflow)
struct st_flow *pflow;
{
	mode_t mode;

	pflow->fd = INVALID_FD;
	if (!*pflow->fullname) GOTO_ERROR;

	if (!gendirs(pflow->fullname, NON)) {
		NL_ERR(0)("Can't create path for file %s\n", pflow->fullname);
		GOTO_ERROR;
	}
	switch (pflow->type) {
	case OPB_BLKDEV: mode = S_IFBLK; goto create_special;
	case OPB_CHRDEV: mode = S_IFCHR; goto create_special;
	case OPB_FIFO:	 mode = S_IFIFO; goto create_special;
	case OPB_DIR:
		if (!(pflow->ark_mode & ~S_IFMT)) pflow->ark_mode = 0755;
		if (mkdir(pflow->fullname, pflow->ark_mode) < 0) {
			if (errno == EEXIST) goto dir_ok ; /* File exists */
			NL_ERR(0)("Can't create directory %s\n",
					pflow->fullname);
			GOTO_ERROR;
		}
dir_ok:
		chown(pflow->fullname, pflow->ark_uid, pflow->ark_gid);
		break ;
	case OPB_NAMEDFILE:	/* Try to restore if there is data ??? */
	case OPB_OBJECT:	/* We create a file with object data */
	case OPB_RAWDEV:	/* We create a file with raw data */
	case OPB_REGFILE:
		if (pflow->flags & HARD_LINK) {		/* Case of hard link */
			if (link(pflow->link_path, pflow->fullname) < 0) {
				NL_ERR(0)("Can't link %s to %s\n",
					  pflow->fullname, pflow->link_path);
				GOTO_ERROR;
			}
			goto ret_oui;
		}
		if (!(pflow->ark_mode & ~S_IFMT)) pflow->ark_mode = 0644;
		if ((pflow->fd = open(pflow->fullname,
					(O_WRONLY | O_CREAT | O_TRUNC),
					 pflow->ark_mode)) < 0) {
			NL_ERR(0)("Can't create file %s\n", pflow->fullname);
			GOTO_ERROR;
		}
		break;
	case OPB_SYMLINK:
		if (!(pflow->ark_mode & ~S_IFMT)) pflow->ark_mode = 0644;
		if (symlink(pflow->link_path, pflow->fullname) < 0) {
			if (errno == EEXIST) unlink(pflow->fullname);
			if (symlink(pflow->link_path, pflow->fullname) < 0) {
				NL_ERR(0)("Can't link %s to %s\n",
					  pflow->fullname, pflow->link_path);
				GOTO_ERROR;
			}
		}
		set_time(pflow->fullname, pflow->ark_atime, pflow->ark_mtime);
		chmod(pflow->fullname, pflow->ark_mode);
		chown(pflow->fullname, pflow->ark_uid, pflow->ark_gid);
		break;
	default:
		NL_ERR(0)("Unknown file type for file %s\n", pflow->fullname);
		GOTO_ERROR;
	}
ret_oui:
	return (OUI);

create_special:
	if (!(pflow->ark_mode & ~S_IFMT)) pflow->ark_mode = 0644;
	if (mknod(pflow->fullname, mode, pflow->ark_rdev) < 0) {
		NL_ERR(0)("Can't create special file %s.\n", pflow->fullname);
		GOTO_ERROR;
	}
	set_time(pflow->fullname, pflow->ark_atime, pflow->ark_mtime);
	chmod(pflow->fullname, pflow->ark_mode);
	chown(pflow->fullname, pflow->ark_uid, pflow->ark_gid);
	return (OUI);
error:
	pflow->operation = DONT_EXTRACT;
	return (NON);
}

/*--------------------------------------------------------------------------*/

int write_file(pflow)
struct st_flow *pflow;
{
	if (!pflow->buff) return(OUI);
	if (pflow->fd == INVALID_FD) {
		if (!create_file(pflow)) GOTO_ERROR;
	}
	if (pflow->flags & COMPRESS_BUFF) uncompress(pflow);
	write(pflow->fd, pflow->buff, pflow->buff_size);
	if (!(pflow->flags & DONT_FREE_BUFF)) free(pflow->buff);
	pflow->buff	 = 0;	/* Don't refer to buffer */
	pflow->buff_size = 0;
	pflow->cur_pc	 = 0;
	return (OUI);
error:
	pflow->operation = DONT_EXTRACT;
	if (!(pflow->flags & DONT_FREE_BUFF)) free(pflow->buff);
	pflow->buff	 = 0;
	pflow->buff_size = 0;
	pflow->cur_pc	 = 0;
	return (NON);
}

/*--------------------------------------------------------------------------*/

static char */*NL_DCL*/getword(line, word, line_end)
char *line;
char *word;
char *line_end;
{
	char *pc1, *pc2;

	for (pc1 = line, pc2 = word;
	     *pc1 && (pc1 < line_end);
	     *pc2 = *pc1, pc1++, pc2++) ;
	*pc2 = 0;
	if (pc1 == line_end) return (0);
	else return (pc1);
}

/*--------------------------------------------------------------------------*/

static char */*NL_DCL*/getnchars(line, word, line_end, n)
char *line;
char *word;
char *line_end;
int n;
{
	char *pc1, *pc2;
	int i;

	for (pc1 = line, pc2 = word, i = n;
	     i > 0 && *pc1 && (pc1 < line_end);
	     *pc2 = *pc1, pc1++, pc2++, i--);
	*pc2 = 0;
	if (pc1 == line_end) return (0);
	else return (pc1);
}

/*---------------------------------------------------------------------------*
 * read next block : returns the number of bytes read.
 *---------------------------------------------------------------------------*/

static int read_block(fd)
int fd;
{
	char *pc = 0;
	int blk_size;
	static int cur_seg = 0;

	blk_size = 0;
	pc	 = (char *)in_blk;
	if (read(fd, pc, TAPEPAGESIZE) <= 0) {
		if (errno == ENOENT) errno = 0;
		goto end_of_segment;
	}
	if (need_uint4_chsex) {
		in_blk->blkseqnum = ch_uint4_sex(in_blk->blkseqnum);
		in_blk->blkbkid	  = ch_uint4_sex(in_blk->blkbkid);
	}
	blk_size = (in_blk->blksz & BLKSZMASK) * TAPEPAGESIZE;
	if (blk_size) {
		pc += TAPEPAGESIZE;
		if (read(fd, pc, blk_size) < blk_size) {
			if (errno == ENOENT) errno = 0;
			goto end_of_segment;
		}
	}
	blk_size += TAPEPAGESIZE;
	return (blk_size);
end_of_segment:
	NL_LOG(10)("read_block: Got end of segment '%d'\n", cur_seg);
	cur_seg++;
	return (0);
}

/*--------------------------------------------------------------------------*/

void get_string_vals(pflow)
struct st_flow *pflow;
{
	char str[STR_SIZE];

	NL_LOG(80)("[%d] Prop '%s'='%s'\n",
		   pflow->fnum, pflow->prop, pflow->value);

	if (!strcmp(pflow->prop, "s_abxvz_ftype")) {
		if (!strcmp(pflow->value, "S_IFOBJ"))	  pflow->type = OPB_OBJECT;
		else if(!strcmp(pflow->value, "S_IFDEV")) pflow->type = OPB_RAWDEV;
		else if(!strcmp(pflow->value, "S_IFDIR")) pflow->type = OPB_DIR;
		else if(!strcmp(pflow->value, "S_IFCHR")) pflow->type = OPB_CHRDEV;
		else if(!strcmp(pflow->value, "S_IFBLK")) pflow->type = OPB_BLKDEV;
		else if(!strcmp(pflow->value, "S_IFREG")) pflow->type = OPB_REGFILE;
		else if(!strcmp(pflow->value, "S_IFIFO")) pflow->type = OPB_FIFO;
		else if(!strcmp(pflow->value, "S_IFNAM")) pflow->type = OPB_NAMEDFILE;
		else if(!strcmp(pflow->value, "S_IFLNK")) pflow->type = OPB_SYMLINK;
		else {
			NL_ERR(0)("Unknown file type: %s\n", pflow->value);
			pflow->type = OPB_UNKNOWN;
		}
	} else if (!strcmp(pflow->prop, "s_host")) {
		strcpy(pflow->host, pflow->value);
		*(pflow->volume) = 0;
		*(pflow->location) = 0;
		if ((global_operation == ADMIN_INFO) &&
		    pflow->slave_version && *pflow->slave_version) {
			INFO(0)("[%d] Got file(s) from slave '%s' version %s\n",
				pflow->fnum, pflow->host, pflow->slave_version);
			*pflow->slave_version = 0;
		}
	} else if (!strcmp(pflow->prop, "s_volum")) {
		strcpy(pflow->volume, pflow->value);
	} else if (!strcmp(pflow->prop, "s_pluga")) {
		strcpy(pflow->plugin, pflow->value);
	} else if (!strcmp(pflow->prop, "s_locat")) {
		strcpy(pflow->location, pflow->value);
	} else if (!strcmp(pflow->prop, "s_cmpnt")) {
		strcpy(pflow->component, pflow->value);
	} else if (!strcmp(pflow->prop, "s_fslnk")) {
		strcpy(pflow->link_path, pflow->value);
	} else if (!strcmp(pflow->prop, "s_flink")) {
		sprintf(str, "%s%s", pflow->host, pflow->value);
		strcpy(pflow->link_path, str);
	} else if (!strcmp(pflow->prop, "s_typeobj")) {
		strcpy(pflow->obj_type, pflow->value);
	}
}

/*--------------------------------------------------------------------------*/

void get_mode_str(pflow, str)
struct st_flow *pflow;
char *str;
{
	mode_t mode;
	char *pc;

	pc = str;
	switch(pflow->type) {
	case OPB_BLKDEV:  *pc = 'b'; break;
	case OPB_CHRDEV:  *pc = 'c'; break;
	case OPB_FIFO:	  *pc = 'p'; break;
	case OPB_DIR:	  *pc = 'd'; break;
	case OPB_SYMLINK: *pc = 'l'; break;
	case OPB_OBJECT:  *pc = 'O'; break;
	case OPB_RAWDEV:  *pc = 'R'; break;
	default:	  *pc = '-'; break;
	}
	if (pflow->type == OPB_DIR && !(pflow->ark_mode & ~S_IFMT)) pflow->ark_mode = 0755;
	else if(!(pflow->ark_mode & ~S_IFMT)) pflow->ark_mode = 0644;
	pc++;

	mode = pflow->ark_mode;
	if (mode & S_IREAD) *pc = 'r';
	else *pc = '-';
	pc++;
	if (mode & S_IWRITE) *pc = 'w';
	else *pc = '-';
	pc++ ;
	if ((mode & S_ISUID) && (mode & S_IEXEC)) *pc = 's';
	else if (mode & S_IEXEC) *pc = 'x';
	else if (mode & S_ISUID) *pc = 'S';
	else *pc = '-';
	pc++ ;

	if (mode & (S_IREAD>>3)) *pc = 'r';
	else *pc = '-';
	pc++;
	if (mode & (S_IWRITE>>3)) *pc = 'w';
	else *pc = '-';
	pc++;
	if ((mode & S_ISGID) && (mode & (S_IEXEC>>3))) *pc = 's';
	else if (mode & (S_IEXEC>>3)) *pc = 'x';
	else if (mode & S_ISGID) *pc = 'l';
	else *pc = '-';
	pc++ ;

	if (mode & (S_IREAD>>6)) *pc = 'r';
	else *pc = '-';
	pc++;
	if (mode & (S_IWRITE>>6)) *pc = 'w';
	else *pc = '-';
	pc++;
	if ((mode & S_ISVTX) && (mode & (S_IEXEC>>6))) *pc = 't';
	else if (mode & (S_IEXEC>>6)) *pc = 'x';
	else if (mode & S_ISVTX) *pc = 'T';
	else *pc = '-';
	pc++;
	*pc = 0;
}

/*--------------------------------------------------------------------------*/

void get_num_vals(pflow)
struct st_flow *pflow;
{
	char *pc;
	char mode_str[STR_SIZE];
	char file_str[STR_SIZE];
	int num;

	num = atoi(pflow->value);
	NL_LOG(80)("[%d] Prop '%s'='%d'\n", pflow->fnum, pflow->prop, num);

	if (!strcmp(pflow->prop, "n_dev"))	 pflow->ark_dev	  = num;
	else if (!strcmp(pflow->prop, "n_ino"))	 pflow->ark_ino	  = num;
	else if (!strcmp(pflow->prop, "n_mode"))	 pflow->ark_mode  = num;
	else if (!strcmp(pflow->prop, "n_nlink")) pflow->ark_nlink = num;
	else if (!strcmp(pflow->prop, "n_uid"))	 pflow->ark_uid	  = num;
	else if (!strcmp(pflow->prop, "n_gid"))	 pflow->ark_gid	  = num;
	else if (!strcmp(pflow->prop, "n_rdev"))	 pflow->ark_rdev  = num;
	else if (!strcmp(pflow->prop, "n_fsize")) pflow->ark_size  = num;
	else if (!strcmp(pflow->prop, "n_atime")) pflow->ark_atime = num;
	else if (!strcmp(pflow->prop, "n_mtime")) pflow->ark_mtime = num;
	else if (!strcmp(pflow->prop, "n_ctime")) pflow->ark_ctime = num;
	else if (!strcmp(pflow->prop, "n_cksum")) {
		if (pflow->operation == NO_OPERATION) {
			/* No data comming with this file */
			check_operation(pflow);
			if (pflow->type == OPB_REGFILE) {
				if (pflow->ark_nlink > 1) pflow->flags |= HARD_LINK;
			}
			if (pflow->operation == EXTRACT_FILE) create_file(pflow);
		}
		else if (pflow->operation == EXTRACT_FILE) {
			NL_LOG(10)("[%d] x %s\n", pflow->fnum, pflow->fullname);
		}
		else if (pflow->operation == CONSULT_LIST) {
			get_mode_str(pflow, mode_str);
			pc = ctime(&pflow->ark_mtime);
			if (pflow->type == OPB_SYMLINK) {
				sprintf(file_str, "%s symbolic link to %s",
					pflow->fullname, pflow->link_path) ;
			} else {
				if (pflow->flags & HARD_LINK) {
					sprintf(file_str, "%s linked to %s",
						pflow->fullname, pflow->link_path) ;
				} else sprintf(file_str,"%s", pflow->fullname);
			}
			NL_LOG(10)("[%d] %s %4d/%-4d%7ld %-12.12s %-4.4s %s\n",
				   pflow->fnum, mode_str, pflow->ark_uid,
				   pflow->ark_gid, (long)pflow->ark_size, pc+4, 
				   pc+20, file_str) ;
		}
		if (pflow->operation == DONT_EXTRACT) {
			NL_LOG(30)("[%d] s %s\n", pflow->fnum, pflow->fullname);
		}
		end_of_file(pflow);
	}
}

/*--------------------------------------------------------------------------*/

void end_of_file(pflow)
struct st_flow *pflow;
{
	UINT4 ui1, ui2;
	char *p1, *p2, *p3, *p4, *p5, *p6, *p7, *p8, *p9, *p10, *p11, *p12;
	int n1;

	if (pflow->operation == EXTRACT_FILE) {
		if (pflow->buff) write_file(pflow);
		if (pflow->fd >= 0) {
			close(pflow->fd);
			set_time(pflow->fullname, pflow->ark_atime,
				 pflow->ark_mtime);
			if (!(pflow->ark_mode & ~S_IFMT)) pflow->ark_mode = 0644;
			chmod(pflow->fullname, pflow->ark_mode);
			chown(pflow->fullname, pflow->ark_uid, pflow->ark_gid);
			pflow->fd = INVALID_FD;
		}
	}

	/* Initialise flow vars */
	p1 = pflow->prop;
	p2 = pflow->value;
	p3 = pflow->realsize;
	p4 = pflow->fullname;
	p5 = pflow->host;
	p6 = pflow->volume;
	p7 = pflow->location;
	p8 = pflow->component;
	p9 = pflow->link_path;
	p10 = pflow->obj_type;
	p11 = pflow->slave_version;
	p12 = pflow->plugin;
	if (p1) *p1 = *p2 = *p3 = *p4 = *p8 = *p9 = *p10 = *p11 = *p12 = 0;

	n1 = pflow->fnum;
	ui1 = pflow->bkid;
	ui2 = pflow->blknum;

	memset(pflow, '\0', sizeof(struct st_flow));

	pflow->prop = p1;
	pflow->value = p2;
	pflow->realsize	= p3;
	pflow->fullname = p4;
	pflow->host = p5;
	pflow->volume = p6;
	pflow->location = p7;
	pflow->component = p8;
	pflow->link_path = p9;
	pflow->obj_type = p10;
	pflow->slave_version = p11;
	pflow->plugin = p12;

	pflow->fnum = n1;
	pflow->bkid = ui1;
	pflow->blknum = ui2;

	pflow->state = NORMAL_STATE;
	pflow->operation = NO_OPERATION;
	pflow->fd = INVALID_FD;
}

/*--------------------------------------------------------------------------*/

char *perform_empty_block(pflow, pc)
struct st_flow *pflow;
char *pc;
{
	int realsize;

	realsize = atoi(pflow->realsize);

	NL_LOG(70)("[%d] Empty block prop '%s' (%d)\n",
		   pflow->fnum, pflow->prop, realsize);
	if (pflow->operation == EXTRACT_FILE) {
		if (pflow->buff) {
			if (!write_file(pflow)) goto ret_oui;
		} else {
			if (pflow->fd == INVALID_FD) {
				if (!create_file(pflow)) goto ret_oui;
			}
		}
		if (realsize) {
			if (lseek(pflow->fd, (off_t)(realsize-1), SEEK_CUR)<0) {
				NL_ERR(0)("Can't seek for free block\n");
				myexit(4);
			}
			write(pflow->fd, "\0", 1);
		}
		goto ret_oui;
	}
ret_oui:
	/* If don't extract forget this empty block */
	pflow->state = NORMAL_STATE;
	return(pc);
}

/*--------------------------------------------------------------------------*/

char *get_bin_vals(pflow, pc, max_pc)
struct st_flow *pflow;
char *pc;
char *max_pc;
{
	char *pc1;
	int data_len, realsize, max_len, nb_bytes;

	data_len = pflow->bin_rest;
	realsize = atoi(pflow->realsize);

	if (pflow->operation == NO_OPERATION) check_operation(pflow);

	if (!strcmp(pflow->prop, "b_9f")) return (perform_empty_block(pflow, pc));

	if (pflow->operation == EXTRACT_FILE) { /* File data */
		if (strlen(pflow->prop) == 4) { /* Block mode */
			if (pflow->state == INCOMPLETE_BIN) {
				NL_LOG(70)("[%d] Continuing Binary prop '%s' (%d)\n",
					   pflow->fnum, pflow->prop, data_len) ;
			} else {
				if (pflow->buff) {
					if (!write_file(pflow)) {
						pflow->operation = DONT_EXTRACT;
						goto skip_data;
					}
				}
				if (realsize &&
				    ((hextoi(pflow->prop + 2) & COMPRESS_MASK) > 0)) {
					pflow->flags |= COMPRESS_BUFF;
					NL_LOG(70)
					  ("[%d] %s Compress Bin prop '%s' (size=%d) (realsize=%d)\n",
					   pflow->fnum, compress_type(pflow->prop),
					   pflow->prop, data_len, realsize);
				} else {
					NL_LOG(70)("[%d] Binary prop '%s' (size=%d)\n",
						   pflow->fnum, pflow->prop, data_len);
				}
				if (!(pflow->buff = malloc(data_len))) {
					NL_ERR(0)("Can't alloc memory for a file buffer\n");
					myexit(4);
				}
				pflow->buff_size = data_len;
				pflow->cur_pc = pflow->buff;
			}
			for (pc1 = pflow->cur_pc;
			     (pc < max_pc) && (data_len > 0);
			     *pc1 = *pc, pc++, pc1++, data_len--) ;
			if (data_len) {
				pflow->cur_pc = pc1;
				pflow->state = INCOMPLETE_BIN;
				pflow->bin_rest = data_len;
				goto unexp_eob;
			} else write_file(pflow);
			goto ret_oui;
		} else if (!strcmp(pflow->prop, "b_fsiz")) {
			if (pflow->fd == INVALID_FD) {
				if (!create_file(pflow)) {
					pflow->operation = DONT_EXTRACT;
					goto skip_data;
				}
			}
			max_len = max_pc - pc;
			nb_bytes = (max_len > data_len) ? data_len : max_len;

			write(pflow->fd, pc, nb_bytes);

			data_len -= nb_bytes;
			pc += nb_bytes;

			if (data_len) {
				pflow->state = INCOMPLETE_BIN;
				pflow->bin_rest = data_len;
				goto unexp_eob;
			}
		} else goto skip_data;
	} else {
		/* Skip those binary data */
		if (pflow->state == INCOMPLETE_BIN) {
			NL_LOG(70)("[%d] Continuing to skip Binary prop '%s' (%d)\n",
				   pflow->fnum, pflow->prop, data_len);
		} else {
skip_data:
			if ((strlen(pflow->prop) == 4) && realsize && 
			   ((hextoi(pflow->prop + 2) & COMPRESS_MASK) > 0)) {
				pflow->flags |= COMPRESS_BUFF;
				NL_LOG(70)
				  ("[%d] Skipping %s compress bin prop '%s' (size=%d) (realsize=%d)\n",
				   pflow->fnum, compress_type(pflow->prop),
				   pflow->prop, data_len, realsize);
			} else {
				NL_LOG(70)("[%d] Skipping binary prop '%s' (size=%d)\n",
					   pflow->fnum, pflow->prop, data_len);
			}
		}
		max_len = max_pc - pc;
		nb_bytes = (max_len > data_len) ? data_len : max_len;

		write(pflow->fd, pc, nb_bytes);

		data_len -= nb_bytes;
		pc += nb_bytes;

		if (data_len) {
			pflow->state = INCOMPLETE_BIN;
			pflow->bin_rest = data_len;
			goto unexp_eob;
		}
	}
ret_oui:
	pflow->state = NORMAL_STATE;
	return (pc);
unexp_eob:
	return (0);
}

/*--------------------------------------------------------------------------*/

char *parse_value(pflow, pc, max_pc)
struct st_flow *pflow;
char *pc;
char *max_pc;
{
	pflow->state = INCOMPLETE_VAL;
	switch (*pflow->prop) {
	case 's':	/* String */
		GETWORD(pc, pflow->value, max_pc);
		get_string_vals(pflow);
		break;
	case 'n':	/* Number */
		GETWORD(pc, pflow->value, max_pc);
		get_num_vals(pflow);
		break;
	case 'b':
		GETWORD(pc, pflow->value, max_pc);	/* Block size */
		pflow->bin_rest = atoi(pflow->value);
		if (strlen(pflow->prop) == 4) {
			pflow->state = INCOMPLETE_REALSZ; /* Real size */
			GETNCHARS(pc, pflow->realsize, max_pc, 5);
			pflow->bin_rest -= 5;
		}
		if (!(pc = get_bin_vals(pflow, pc, max_pc))) goto unexp_eob;
		break ;
	}
	pflow->state = NORMAL_STATE;
	return (pc);
unexp_eob:
	return (0);
}

/*--------------------------------------------------------------------------*/

char *complete_missing(pflow, pc, max_pc)
struct st_flow *pflow;
char *pc;
char *max_pc;
{
	char str[STR_SIZE];
	int n;

	if (pflow->state == INCOMPLETE_PROP) {
		GETWORD(pc, str, max_pc);
		strcat(pflow->prop, str);
		if (*pflow->prop == 'b') {
			GETWORD(pc, pflow->value, max_pc); /* Block size */
			pflow->bin_rest = atoi(pflow->value);
			if (strlen(pflow->prop) == 4) {
				pflow->state = INCOMPLETE_REALSZ;
				GETNCHARS(pc, pflow->realsize, max_pc, 5); /* Real size */
				pflow->bin_rest -= 5;
			}
			return (get_bin_vals(pflow, pc, max_pc));
		}
		if (!(pc = parse_value(pflow, pc, max_pc))) goto unexp_eob;
		return (pc);
	}
	if (pflow->state == INCOMPLETE_VAL) {
		GETWORD(pc, str, max_pc);
		strcat(pflow->value, str);
		if (*pflow->prop == 'b') {
			pflow->bin_rest = atoi(pflow->value);
			if (strlen(pflow->prop) == 4) {
				GETNCHARS(pc, pflow->realsize, max_pc, 5); /* Real size */
				pflow->bin_rest -= 5;
			}
			return (get_bin_vals(pflow, pc, max_pc));
		}
		strcpy(str, pflow->value);
		(void) parse_value(pflow, str, &str[STR_SIZE]);
		return (pc);
	}
	if (pflow->state == INCOMPLETE_REALSZ) {
		n = strlen(pflow->realsize);
		GETNCHARS(pc, str, max_pc, 5-n);
		strcat(pflow->realsize, str);
		pflow->bin_rest = atoi(pflow->value) - 5;
		return (get_bin_vals(pflow, pc, max_pc));
	}
	if (pflow->state == INCOMPLETE_BIN) {
		return (get_bin_vals(pflow, pc, max_pc));
	}
	return (pc);
unexp_eob:
	return (NON);
}

/*--------------------------------------------------------------------------*/

int parse_block(pflow, pblk, blk_size)
struct st_flow *pflow;
struct stapblk *pblk;
int blk_size;
{
	char *pc, *max_pc;

	max_pc = ((char *)pblk) + blk_size;

	if (pflow->state == NORMAL_STATE) {
		pc =((char *)pblk) + ((pblk->blkproffh << 8) | pblk->blkproffl);
	} else {
		pc = ((char *)pblk) + HEADER_SIZE;
		if (!(pc = complete_missing(pflow, pc, max_pc))) goto unexp_eob;
	}

	pflow->state = NORMAL_STATE;
	while (*pc && (pc < max_pc)) {
		pflow->state = INCOMPLETE_PROP;
		GETWORD(pc, pflow->prop, max_pc);
		if (!strcmp(pflow->prop, OPB_SERV_VERSION)) {
			GETWORD(pc, pflow->value, max_pc);
			strcpy(pflow->slave_version, pflow->value);
			pflow->state = NORMAL_STATE;
			continue ;
		}
		if (!(pc = parse_value(pflow, pc, max_pc))) goto unexp_eob;
	}
	return (OUI);
unexp_eob:
	NL_LOG(40)("[%d] We need another block to complete.\n", pflow->fnum);
	return (OUI);
}

/*--------------------------------------------------------------------------*/

int read_label(fd)
int fd;
{
	int offset;
	char *pc, *max_pc;
	char prop[STR_SIZE];
	char value[STR_SIZE];
	int label_size = 0;
	int num;
	struct stapblk *pblk=0;
	char tapeid[STR_SIZE];
	char author[STR_SIZE];
	char label_name[STR_SIZE];

	if (!(label_size = read_block(fd))) GOTO_ERROR;
	pblk = in_blk;
	if (pblk->blkseqnum != O3MAGICTAPE) {
		if (ch_uint4_sex(pblk->blkseqnum) != O3MAGICTAPE) {
			NL_ERR(0)("This tape is not a ARKEIA tape.\n");
			GOTO_ERROR;
		}
	}
	offset = ((pblk->blkproffh << 8) | pblk->blkproffl);
	pc = ((char *)pblk) + offset;
	max_pc = ((char *)pblk) + label_size;

	while (*pc && (pc < max_pc)) {
		GETWORD(pc, prop, max_pc);
		if (!strcmp(prop, OPB_VERSION)) {
			GETWORD(pc, value, max_pc);
			INFO(0)("Server version %s\n", value);
			continue ;
		}
		switch (*prop) {
		case 's':	/* String */
			GETWORD(pc, value, max_pc);
			NL_LOG(80)("Prop '%s'='%s'\n", prop, value);
			if (!strcmp(prop, "s_lnam")) strcpy(label_name, value);
			else if (!strcmp(prop, "s_auth")) strcpy(author, value);
			break ;
		case 'n':
			GETWORD(pc, value, max_pc);
			num = atoi(value);
			NL_LOG(80)("Prop '%s'='%d'\n", prop, num);
			if (!strcmp(prop, "n_tapeid")) strcpy(tapeid, value);
			break ;
		}
	}
	if (global_operation == ADMIN_INFO) {
		INFO(0)("Reading tape '%s' created by %s (tapeid=%s)\n",
			label_name, author, tapeid);
	} else if(global_operation == EXTRACT_FILE) {
		NL_LOG(10)("Extracting data from tape '%s' created by %s\n",
			   label_name, author);
	} else {
		NL_LOG(10)("Consulting data from tape '%s' created by %s\n",
			   label_name, author);
	}
	return(label_size);
error:
	NL_ERR(0)("Can't read tape label.\n");
	return (0);
unexp_eob:
	NL_ERR(0)("Found unexpected end of block.\n");
	return (0);
}

/*--------------------------------------------------------------------------*/

#define MBUFSIZE	4096		/* 4K buffer */

int read_tape(device_name, label_check)
char *device_name;
int label_check;
{
	char *buf ;
	char temp_file[STR_SIZE] ;
	int blk_size ;
	int blknum, flow_num ;
	int delete ;
	int n ;
	int nb_null ;
	int nb_retry ;
	int tmp_fd ;
	struct st_flow *pflow ;

	if(global_operation == ADMIN_INFO) {
		INFO(0)("Using readarkeia version '%s'\n", readarkeia_version) ;
	}

	delete = NON ;
	if(!strcmp(device_name, "-")) {
		sprintf(temp_file, "/usr/tmp/readarkeia_%d", getpid()) ;
		tmp_fd = open(temp_file,
				O_RDWR | O_CREAT | O_TRUNC | O_EXCL, 0644) ;
		if(tmp_fd < 0) {
			NL_ERR(0)("readarkeia: temporary file %s\n", temp_file);
			myexit(2);
		}
		buf = malloc(MBUFSIZE);
		if (!buf) {
			NL_ERR(0)("tar: Out of memory\n");
			myexit(2);
		}
		while ((n = read(0, buf, MBUFSIZE)) > 0) {
			write(tmp_fd, buf, n);
		}
		free(buf);
		close(tmp_fd);
		device_name = temp_file;
		delete = OUI;
	}
	nb_retry = 0;
retry:
	if ((fd_in = open(device_name, O_RDONLY))<0) {
		if (errno == EAGAIN) {
			if (nb_retry < 5) {
				NL_LOG(40)
				  ("Resource temporarily unavailable. Retrying...\n");
				nb_retry++;
				sleep(2);
				goto retry;
			}
		}
		NL_ERR(0)("Can't open device %s\n", device_name);
		GOTO_ERROR;
	}
	if (!(in_blk = (struct stapblk *)malloc(64 * 1024))) {
		NL_ERR(0)("Can't alloc memory for label blk\n");
		GOTO_ERROR;
	}
	if (label_check && !read_label(fd_in)) GOTO_ERROR;

	nb_null = 0;
	while (1) {
		if (nb_null > 3) {
			NL_LOG(40)("End of tape.\n");
			goto ret_oui;
		}
		if (!(blk_size = read_block(fd_in))) {
			nb_null++;
			continue ;
		}
		nb_null = 0;
		flow_num = COMP2FLOW(in_blk->blkseqnum);
		if (flow_num == F_ST) {
			NL_LOG(40)("Got Strategy process block. skipping...\n");
			continue ;
		}
		if (in_blk->blksz & DB_BLOCK) {
			NL_LOG(40)("Got database block. skipping...\n");
			continue ;
		}

		/* It's a data block */
		if ((flow_num < 0) || (flow_num >= MAXFLOWS)) {
			NL_ERR(0)
			  ("Found a block with a flow number higher than %d: %d\n",
			   MAXFLOWS, flow_num);
			myexit(9);
		}
		blknum = (in_blk->blkseqnum & 0xffffff);
		pflow = &flow_tab[flow_num-1];
		if (blknum != (pflow->blknum+1)) { /* It's another backup */
			if ((pflow->state != NORMAL_STATE) ||
			    (pflow->buff)) {
				NL_LOG(10)
				  ("[%d] Warning file %s has been damaged.\n",
				   pflow->fnum, pflow->fullname);
				if (pflow->operation == EXTRACT_FILE) {
					NL_LOG(10)("[%d] Extracting new file.\n",
						   pflow->fnum);
				}
			}
			end_of_file(pflow);
		} else {
			if (blknum <= pflow->blknum) { /* Got duplicate block */
				NL_LOG(40)
				  ("Got an already read block %d for flow %d. skipping...\n",
				   blknum, pflow->fnum);
				continue ;
			}
		}
		NL_LOG(40)("[%d] Start of Block %d, size %d, bkid %d.\n",
			   pflow->fnum, blknum, blk_size, in_blk->blkbkid) ;

		/* Allocate strings needed by this flow is they are not already */
		ALLOC_IF_NOT(pflow->prop);
		ALLOC_IF_NOT(pflow->value);
		ALLOC_IF_NOT(pflow->realsize);
		ALLOC_IF_NOT(pflow->fullname);
		ALLOC_IF_NOT(pflow->host);
		ALLOC_IF_NOT(pflow->plugin);
		ALLOC_IF_NOT(pflow->volume);
		ALLOC_IF_NOT(pflow->location);
		ALLOC_IF_NOT(pflow->component);
		ALLOC_IF_NOT(pflow->link_path);
		ALLOC_IF_NOT(pflow->obj_type);
		ALLOC_IF_NOT(pflow->slave_version);

		if (pflow->bkid && (pflow->bkid != in_blk->blkbkid)) {
			NL_ERR(0)
			  ("Got another backup id for flow %d: %d --> %d\n",
			   pflow->fnum, pflow->bkid, in_blk->blkbkid);
		}
		pflow->blknum = blknum;
		if (!parse_block(pflow, in_blk, blk_size)) GOTO_ERROR;
	}
ret_oui:
	if (fd_in >= 0) {
		CLOSE(fd_in);
		fd_in = INVALID_FD;
	}
	if (delete) unlink(device_name);
	return (OUI);
error:
	if (fd_in >= 0) {
		CLOSE(fd_in);
		fd_in = INVALID_FD;
	}
	if (delete) unlink(device_name);
	return (NON);
}

/*--------------------------------------------------------------------------*/

int main(argc, argv)
int argc;
char **argv;
{
	char *pc;
	char c;
	char device_name[STR_SIZE];
	int i;
	int inc;
	int label_check;

	argv++;
	if (!--argc) usage();
	if (argv[0][0] == '-') {
		switch (argv[0][1]) {
		case 'l':
			nl_loglevel = atoi(&argv[0][2]);
			break ;
		}
		argc--;
		argv++;
	}
	if (!argc) usage();
	inc = 1;
	label_check = OUI;
	for (i = 0 ; argv[0][i] ; i++) {
		switch (argv[0][i]) {
		case 'i':
			if (global_operation != NO_OPERATION) usage();
			global_operation = ADMIN_INFO;
			break ;
		case 'f':
			if (argc < 2) usage();
			strcpy(device_name, argv[1]);
			inc++;
			break ;
		case 'm':
			label_check = NON;
			break ;
		case 't':
			if (global_operation != NO_OPERATION) usage();
			global_operation = CONSULT_LIST;
			break ;
		case 'v':
			if (!nl_loglevel) nl_loglevel = 10;
				/* First level of verbose */
			break ;
		case 'x':
			if (global_operation != NO_OPERATION) usage();
			global_operation = EXTRACT_FILE;
			break ;
		case 'u':
			gl_update_mode = OUI;
			break;
		default:
			usage();
		}
	}
	argc -= inc;
	argv += inc;
	while (argc) {
		i = strlen(argv[0]);
		pc = malloc(i+1);
		strcpy(pc, argv[0]);
		extract_path[nb_path] = pc;
		nb_path++;
		argc--;
		argv++;
	}
	if (!*device_name) usage();
	INFO(0)("Device name = %s\n", device_name);
	for (i = 0; i < MAXFLOWS; i++) {
		memset((char *)(&flow_tab[i]), '\0', sizeof(struct st_flow));
		flow_tab[i].fnum = i;
		flow_tab[i].state = NORMAL_STATE;
		flow_tab[i].operation = NO_OPERATION;
		flow_tab[i].fd = INVALID_FD;
	}
	while (1) {
		if (!read_tape(device_name, label_check)) GOTO_ERROR;
		label_check = OUI;
		if (!strcmp(device_name, "-")) goto ret_oui;
		printf("Found end of tape. ");
		printf("Please insert next tape to continue\nand press return or ");
		printf("press 'q' to quit.\n<PRESS ONE KEY>\n");
		fflush(stdout);
		c = getchar();
		if ((c == 'q') || (c == 'Q')) break;
	}
ret_oui:
	myexit(0);
	return (0);
error:
	myexit(3);
	return (3);
}

/*--------------------------------------------------------------------------*/

void myexit(i)
int i;
{
	if (fd_in >= 0) close(fd_in);
	exit(i);
}

/*--------------------------------------------------------------------------*/

void usage()
{
	fprintf(stderr,	"USAGE (V%s):\n\t", readarkeia_version) ;
	fprintf(stderr,
"readarkeia [-l<loglevel>] x|t|i [mfv] [device|-] [filename1 filename2 ...]\n");
	myexit(2);
}

/*--------------------------------------------------------------------------*/

int hextoi(str)
char *str;
{
	register char *pc;
	int i;

	for (i = 0, pc = str; ; pc++) {
		switch (*pc) {
		case 0:
		case '\t':
		case ' ':
			continue ;
		}
		break ;
	}
	for (; ; pc++) {
		if ((*pc >= '0') && (*pc <= '9')) i = i * 16 + (*pc - '0');
		else if ((*pc >= 'a') && (*pc <= 'f')) i = i * 16 + (*pc - 'a' + 10);
		else if ((*pc >= 'A') && (*pc <= 'F')) i = i * 16 + (*pc - 'A' + 10);
		else break ;
	}
	return (i);
}

/*--------------------------------------------------------------------------*
 *---------------------------------UNCOMPRESS LAYER-------------------------*
 *--------------------------------------------------------------------------*/

static char *pwrkmem;
static ULONG wrklen = 0;

typedef void (*void_funcp)();

#define fast_copy(src,dest,len) memcpy(dest,src,len)

#define U(X) ((ULONG) X)

#define FLAG_BYTES	   4
#define FLAG_COPY	   1

/*--------------------------------------------------------------------------*/

char *compress_type(prop)
char *prop;
{
	int mask;

	mask = hextoi(prop + 2);
	if ((mask & COMPRESS_MASK) == LZ1_COMPRESS) return ("lz1");
	else if ((mask & COMPRESS_MASK) == LZ3_3_COMPRESS) return ("lz3_3");
	else return("??");
}

/*--------------------------------------------------------------------------*
 * Based on lzrw1-a.c by Ross Williams
 *--------------------------------------------------------------------------*/

ULONG lz1_memory = U(U(sizeof(long))*U(4096)+U(3));

/*--------------------------------------------------------------------------*/

static void lz1_decompress(wrk_mem, p_src_first, src_len, p_dst_first, p_dst_len)
UBYTE *p_src_first, *p_dst_first;
ULONG src_len, *p_dst_len;
UBYTE *wrk_mem;
{
	register UBYTE *p_src = p_src_first+FLAG_BYTES, *p_dst = p_dst_first;
	UBYTE *p_src_post = p_src_first+src_len;
	UBYTE *p_src_max16 = p_src_first+src_len-(16*2);
	register ULONG control = 1; /* Put 1 in variable control */

	if (*p_src_first == FLAG_COPY) {
		fast_copy(p_src_first + FLAG_BYTES, p_dst_first, src_len - FLAG_BYTES);
		*p_dst_len = src_len-FLAG_BYTES;
		return ;
	}
	while (p_src!=p_src_post) {
		register UWORD unroll;

		if (control == 1) {
			control = 0x10000|*p_src++;
			control |= (*p_src++) << 8;
		}
		unroll =  p_src <= p_src_max16 ? 16 : 1;
		while (unroll--) {
			if (control&1) {
				register UWORD lenmt;
				register UBYTE *p;

				lenmt = *p_src++;
				p = p_dst - (int)(((lenmt&0xF0) << 4) | *p_src++);
				*p_dst++ = *p++;
				*p_dst++ = *p++;
				*p_dst++ = *p++;
				lenmt &= 0xF;
				while (lenmt--) *p_dst++ = *p++;
			} else *p_dst++ = *p_src++;
			control >>= 1;
		}
	}
	*p_dst_len = p_dst-p_dst_first;
}

/*--------------------------------------------------------------------------*
 * Based on lzrw3-a.c by Ross Williams
 *--------------------------------------------------------------------------*/

#define SIZE_P_BYTE	(U(sizeof(UBYTE *)))
#define ALIGNMENT_FUDGE (U(16))
#define MEM_REQ		(U(4096)*(SIZE_P_BYTE) + ALIGNMENT_FUDGE)

ULONG lz3_3_memory = MEM_REQ;

/*--------------------------------------------------------------------------*/

#define HASH_TABLE_DEPTH_BITS		(3)
#define HASH_TABLE_LENGTH		(4096)
#define MAX_CMP_GROUP			(2+16*2)
#define MAX_RAW_GROUP			(16*MAX_RAW_ITEM)
#define MAX_RAW_ITEM			(18)
#define ULONG_ALIGN_UP(X)		((((ULONG)X)+3)&~3)

#define PARTITION_LENGTH_BITS		(12-HASH_TABLE_DEPTH_BITS)
#define PARTITION_LENGTH		(1<<PARTITION_LENGTH_BITS)
#define HASH_TABLE_DEPTH		(1<<HASH_TABLE_DEPTH_BITS )
#define HASH_MASK			(PARTITION_LENGTH-1)
#define DEPTH_MASK			(HASH_TABLE_DEPTH-1)

#define START_STRING_18			((UBYTE *) "123456789012345678")

#define HASH(PTR) \
( \
 (((40543*(((UINT)(*(PTR))<<8)^((UINT)(*((PTR)+1))<<4)^(UINT)(*((PTR)+2))))>>4) & HASH_MASK) \
 << HASH_TABLE_DEPTH_BITS \
 )

#define UPDATE_P(P_BASE,NEWPTR) \
{(P_BASE)[cycle++] = (NEWPTR) ; cycle &= DEPTH_MASK ;}

#define UPDATE_I(I_BASE,NEWPTR) \
{hash[(I_BASE)+cycle++] = (NEWPTR) ; cycle &= DEPTH_MASK ;}

#define ANY_HASH_INDEX (0)

/*--------------------------------------------------------------------------*/

#define ZJ *p_h++ = START_STRING_18

static void lz3_3_decompress(p_wrk_mem, p_src_first, src_len, p_dst_first, p_dst_len)
UBYTE *p_wrk_mem;
UBYTE *p_src_first;
ULONG  src_len;
UBYTE *p_dst_first;
ULONG *p_dst_len;
{
	register UBYTE *p_src = p_src_first+FLAG_BYTES;
	register UBYTE *p_dst = p_dst_first;
	UBYTE *p_src_post  = p_src_first + src_len;
	UBYTE *p_src_max16 = p_src_first + src_len - (MAX_CMP_GROUP - 2);
	UBYTE **hash = (UBYTE **) ULONG_ALIGN_UP(p_wrk_mem);
	register ULONG control = 1;
	register UCARD literals = 0;
	UCARD cycle = 0;

	if (*p_src_first == FLAG_COPY) {
		fast_copy(p_src_first + FLAG_BYTES, p_dst_first, src_len - FLAG_BYTES);
		*p_dst_len = src_len - FLAG_BYTES;
		return ;
	}
	{
		UCARD i;
		UBYTE **p_h = hash;
		for (i = 0; i < 256; i++) {
			ZJ; ZJ; ZJ; ZJ;
			ZJ; ZJ; ZJ; ZJ;
			ZJ; ZJ; ZJ; ZJ;
			ZJ; ZJ; ZJ; ZJ;
		}
	}
	while (p_src != p_src_post) {
		register UCARD unroll;

		if (control == 1) {
			control = 0x10000| * p_src++;
			control |= (*p_src++) << 8;
		}
		unroll = p_src <= p_src_max16 ? 16 : 1;

		while (unroll--) {
			if (control & 1) {
				register UBYTE *p;
				register UCARD lenmt;
				register UBYTE *p_ziv = p_dst;
				register UCARD the_index;

				lenmt = *p_src++;
				the_index = ((lenmt & 0xF0) << 4) | *p_src++;
				p = hash[the_index];
				lenmt &= 0xF;

				*p_dst++ = *p++;
				*p_dst++ = *p++;
				*p_dst++ = *p++;
				while (lenmt--) *p_dst++ = *p++;

				if (literals > 0) {
					register UBYTE *r = p_ziv-literals;
					UPDATE_I(HASH(r), r);
					if (literals == 2) {
						r++;
						UPDATE_I(HASH(r), r);
					}
					literals = 0;
				}
				UPDATE_I(the_index & (~DEPTH_MASK), p_ziv);
			} else {
				*p_dst++ = *p_src++;
				if (++literals == 3) {
					register UBYTE *p = p_dst - 3;
					UPDATE_I(HASH(p), p);
					literals = 2;
				}
			}
			control >>= 1;
		}
	}
	*p_dst_len = p_dst - p_dst_first;
}

/*--------------------------------------------------------------------------*
 * Check working memory required for each type of algorigthm
 *--------------------------------------------------------------------------*/

int check_wrkmem(cmptype)
int cmptype;
{
	switch (cmptype) {
	case LZ1_COMPRESS :
		NL_LOG(40)("Working mem for lz1 algo is %lu\n", lz1_memory);
		if (lz1_memory > wrklen) {
			wrklen = lz1_memory;
			goto realloc_wrkmem;
		}
		break ;
	case LZ3_3_COMPRESS :
		NL_LOG(40)("Working mem for lz3 algo is %lu\n", lz3_3_memory);
		if (lz3_3_memory > wrklen) {
			wrklen = lz3_3_memory;
			goto realloc_wrkmem;
		}
		break ;
	}
	return (OUI);
realloc_wrkmem:
	if (pwrkmem) free(pwrkmem);
	if (!(pwrkmem = malloc(wrklen))) {
		NL_ERR(0)("Can't alloc work buff of %lu bytes for compression\n",
			  wrklen);
		GOTO_ERROR;
	}
	return (OUI);
error:
	return (NON);
}

/*--------------------------------------------------------------------------*/

void uncompress(pflow)
struct st_flow *pflow;
{
	UINT4 maxlen;
	ULONG compsize, realsize, len_uncompress;
	int mask;
	static UINT4 newbuf_size = 0;
	static char *newbuf = 0;
	void_funcp decompress_fn;

	mask	 = hextoi(pflow->prop + 2);
	compsize = atoi(pflow->value) - 5;
	realsize = atoi(pflow->realsize);

	switch(mask & COMPRESS_MASK) {
	case LZ1_COMPRESS:
		if (!check_wrkmem(LZ1_COMPRESS)) GOTO_ERROR;
		decompress_fn = lz1_decompress;
		break ;
	case LZ3_3_COMPRESS:
		if (!check_wrkmem(LZ3_3_COMPRESS)) GOTO_ERROR;
		decompress_fn = lz3_3_decompress;
		break ;
	case NO_COMPRESS :
	default :
		NL_LOG(50)("Not compressed...\n");
		return ;	/* do nothing */
	}
	maxlen = MAX(2 * compsize, realsize + 1024);
	NL_LOG(40)("Assuming memory required of %u\n", maxlen);
	if (newbuf_size < maxlen) {
		if (newbuf) free(newbuf);
		newbuf_size = maxlen;
		if (!(newbuf = malloc((unsigned int)newbuf_size))) {
			NL_ERR(0)("Can't alloc buffer for compression\n");
			newbuf_size = 0;
			GOTO_ERROR;
		}
	}
	decompress_fn(pwrkmem, pflow->buff, compsize, newbuf, &len_uncompress);
	NL_LOG(40)("Uncompressed %ld bytes to %ld\n", compsize, len_uncompress);
	if (len_uncompress != realsize) {
		NL_ERR(0)("WARNING: %ld bytes expand to %ld bytes instead of %ld\n",
			  compsize, len_uncompress, realsize);
		GOTO_ERROR;
	}
	free(pflow->buff);
	pflow->buff = newbuf;
	pflow->buff_size = len_uncompress;
	pflow->flags |= DONT_FREE_BUFF;
	return ;
error:
	return ;	/* Buffer won't be uncompressed */
}

/*-----------------------------------END------------------------------------*/
